#include "DrawWindow.hh"

#include <cstdio>
#include <iostream>
#include <vector>
#include <list>
#include <values.h>
#include <cmath>
#include <unistd.h>
#include <set>
#include <map>

/* Maximum error threshold.
 * Value "0" won't work due to floating point issues.
 * Values between 1e-10 and 1.0 are ok.
 * Values greater than 1 cause visible errors, but give speed.
 */
#define THRESHOLD 0.9

using namespace std;

static vector<bool> imageData;
static vector<unsigned> colorData;
static unsigned xSize, ySize;

static unsigned getcoordnum(unsigned x,unsigned y) { return x + y*xSize; }
static bool imageVal(unsigned x, unsigned y) { return imageData[x+y*xSize]; }
static unsigned colorVal(unsigned x, unsigned y)
{ return colorData[getcoordnum(x,y)]; }
static void setColVal(unsigned x, unsigned y, unsigned val)
{ colorData[x+y*xSize] = val; }


static void setColor(DrawWindow& draw, unsigned color)
{
    switch(color)
    {
      case 0: draw.DrawingColor(0,0,0); break;
      case 1: draw.DrawingColor(65535,32768,65535); break;
      case 2: draw.DrawingColor(65535,65535,32768); break;
      case 3: draw.DrawingColor(65535,65535,65535); break;
      case 4: draw.DrawingColor(20000,40000,20000); break;
      case 5: draw.DrawingColor(10000,10000,50000); break;
    }
}

static void Draw(DrawWindow& draw)
{
    setColor(draw,0);
    draw.Clear();
    setColor(draw,5);

    for(unsigned y=0; y<ySize; y++)
    {
        for(unsigned x=0; x<xSize; x++)
        {
            if(imageVal(x, y)) draw.Point(x, y);
        }
    }
    unsigned col = 5;
    for(unsigned y=0; y<ySize; y++)
    {
        for(unsigned x=0; x<xSize; x++)
        {
            unsigned newcol=colorVal(x,y);
            if(newcol)
            {
                if(col != newcol)
                { col = newcol; setColor(draw, col); }
                draw.Point(x,y);
            }
        }
    }
}

static bool HasEdge(unsigned x,unsigned y)
{
    bool top    = !y, bottom = y==(ySize-1);
    bool left   = !x, right  = x==(xSize-1);
    return (!left             && imageVal(x-1, y  )==false)
        || (!right            && imageVal(x+1, y  )==false)
        || (!top              && imageVal(x  , y-1)==false)
        || (!bottom           && imageVal(x  , y+1)==false)/*
        || (!left  && !top    && imageVal(x-1, y-1)==false)
        || (!left  && !bottom && imageVal(x-1, y+1)==false)
        || (!right && !top    && imageVal(x+1, y-1)==false)
        || (!right && !bottom && imageVal(x+1, y+1)==false)*/;
}

struct path
{
    vector<unsigned> points;
    double cost;
public:
    path() : cost(0) { points.reserve(64); }
    
    /* Cost of going to the next point is defined
     * as the distance between the latest point and
     * the point in question, plus the cost already
     * experienced during this path.
     * If there were different terrain types, this
     * function could indicate them by influencing
     * the cost.
     */
    double checkcost(unsigned nx,unsigned ny) const
    {
        if(!points.size()) return 0;
        int distx = (int)(nx - points[points.size()-1]%xSize);
        int disty = (int)(ny - points[points.size()-1]/xSize);
        return cost + sqrt(double(distx*distx + disty*disty));
    }
    
    void addpoint(unsigned nx,unsigned ny)
    {
        /* Add the cost of going to new point */
        cost = checkcost(nx, ny);
        /* Append the point into the path */
        points.push_back(getcoordnum(nx, ny));
    }
    
    /* Comparator function which compares the cost
     * Needed for automatic sorting by efficiency.
     * Changing its polarity won't mean "finds worst path",
     * but it means it will miss the fastest path.
     */
    inline bool operator< (const struct path &b) const
    {
        return cost < b.cost;
    }
};

/* This function plots a line from (x1,y1) to (x2,y2).
 * It uses the well-known Bresenhamn algorithm which is
 * pretty fast because it avoids any multiplications/divisions.
 */
static void DrawLine(unsigned x1,unsigned y1, unsigned x2,unsigned y2, unsigned color)
{
    unsigned np;
    int dx = x2-x1, dy=y2-y1;
    if(dx<0)dx=-dx; if(dy<0)dy=-dy;
    
    int d, dp1, dp2, yp1,yp2, xp1,xp2;
    if(dx >= dy)
    {
        np = dx;
        d = (dp1 = dy<<1) - dx;
        dp2 = (dy-dx) << 1;
        yp1=0, xp1=xp2=yp2=1;
    }
    else
    {
        np = dy;
        d = (dp1 = dx<<1) - dy;
        dp2 = (dx-dy) << 1;
        xp1=0, yp1=yp2=xp2=1;
    }
    if(x1 > x2) { xp1=-xp1; xp2=-xp2; }
    if(y1 > y2) { yp1=-yp1; yp2=-yp2; }
    unsigned x=x1, y=y1;
    for(unsigned i=0; i<=np; ++i)
    {
        setColVal(x,y, color);
        if(d < 0) d += dp1, x += xp1, y += yp1;
        else      d += dp2, x += xp2, y += yp2;
    }
}

/* A variant of DrawLine that instead of drawing, checks if the route
 * from source to destination is clear.
 */
static bool SeePoint(unsigned x1,unsigned y1, unsigned x2,unsigned y2)
{
    /* Task: Draw a line between the points and
     * check if the points collide with a wall. */
    
    unsigned np;
    int dx = x2-x1, dy=y2-y1;
    if(dx<0)dx=-dx; if(dy<0)dy=-dy;
    
    int d, dp1, dp2, yp1,yp2, xp1,xp2;
    if(dx >= dy)
    {
        np = dx;
        d = (dp1 = dy<<1) - dx;
        dp2 = (dy-dx) << 1;
        yp1=0, xp1=xp2=yp2=1;
    }
    else
    {
        np = dy;
        d = (dp1 = dx<<1) - dy;
        dp2 = (dx-dy) << 1;
        xp1=0, yp1=yp2=xp2=1;
    }
    if(x1 > x2) { xp1=-xp1; xp2=-xp2; }
    if(y1 > y2) { yp1=-yp1; yp2=-yp2; }
    unsigned x=x1, y=y1;
    for(unsigned i=0; i<np; ++i)
    {
        if(imageVal(x,y) == false) return false;
        if(d < 0) d += dp1, x += xp1, y += yp1;
        else      d += dp2, x += xp2, y += yp2;
    }
    return true;
}

static map<unsigned, set<unsigned> > allneighbours;
static set<unsigned> edgepoints;

static void CalcNeighbours(DrawWindow& draw)
{
    typedef set<unsigned>::const_iterator edgeit;
    
    /* Add all points that reside right next to a wall */
    for(unsigned y=0; y<ySize; ++y)
        for(unsigned x=0; x<xSize; ++x)
        {
            if(!HasEdge(x,y) || imageVal(x, y)==false)continue;
            
            // insert this point into the edge list
            edgepoints.insert(getcoordnum(x,y));
            
            // Plot the point just to demonstrate we're doing work
            setColor(draw, 1);
            draw.Point(x,y);
            draw.RefreshArea(x,y, 1,1);
            
            //if(edgepoints.size() >= 300) break;
        }
    
    fprintf(stderr, "%u edge points\n", edgepoints.size());

    /* Find out which points can see which other points */
    for(edgeit i=edgepoints.begin(); i!=edgepoints.end(); ++i)
    {
        unsigned x1 = *i % xSize, y1 = *i / xSize;
        // Find all pairs of points (i and j).
        for(edgeit j=i; ++j!=edgepoints.end(); )
        {
            unsigned x2 = *j % xSize, y2 = *j / xSize;
            
            // Check whether these two points can see each others.
            if(SeePoint(x1,y1, x2,y2))
            {
                // mark i and j as neighbours to each others
                allneighbours[*i].insert(*j);
                allneighbours[*j].insert(*i);
            }
        }
        
        // Unplot the point as we've handled it now
        setColor(draw, 5);
        draw.Point(x1,y1);
        draw.RefreshArea(x1,y1, 1,1);
    }
}

/* Single point add/delete neightbour functions */
static void AddNeighbour(unsigned x,unsigned y)
{
	typedef set<unsigned>::iterator psetit;
	typedef map<unsigned, set<unsigned> >::iterator naapit;

	unsigned coord = getcoordnum(x, y);

	if(edgepoints.find(coord) != edgepoints.end()) return;
	edgepoints.insert(coord);
	
	naapit i = allneighbours.find(coord);
	if(i == allneighbours.end())
	    for(psetit j=edgepoints.begin(); j!=edgepoints.end(); ++j)
	    {
            unsigned x2 = *j % xSize, y2 = *j / xSize;
            if(SeePoint(x,y, x2,y2))
            {
                allneighbours[coord].insert(*j);
                allneighbours[*j].insert(coord);
            }
	    }
}
static void DeleteNeighbour(unsigned x,unsigned y)
{
	typedef set<unsigned>::iterator psetit;
	typedef map<unsigned, set<unsigned> >::iterator naapit;

	if(HasEdge(x,y)) return;
	unsigned coord = getcoordnum(x, y);
	
	edgepoints.erase(coord);
	allneighbours.erase(coord);
	
	for(naapit i = allneighbours.begin(); i != allneighbours.end(); ++i)
		i->second.erase(coord);
}

/* The actual algorithm: finds route. */
static path GetRoute(unsigned x1,unsigned y1, unsigned x2,unsigned y2,
                     DrawWindow& draw)
{
    const double maxcost = double(imageData.size()) * imageData.size();
    
    vector<double> seen(imageData.size(), maxcost);
    
    /* Add the begin and end points to the list of points,
     * just so that actual route can be plotted to/from them. */
    AddNeighbour(x1,y1);
    AddNeighbour(x2,y2);
    
    // We only have this requirement:
    //  - *begin() must always return the smallest value
    // We use multiset. Better ideas, anyone?
    
    multiset<path> paths;          // use multiset for automatical sorting
    
    seen[getcoordnum(x1, y1)] = 0; // never need to return to the first point
    path winner;                   // assume the first try wins ;)
    winner.addpoint(x1, y1);
    paths.insert(winner);
    
    while(!paths.empty())
    {
        /* Pick one of the paths that we're still interested in. */
        multiset<path>::iterator i = paths.begin();

        unsigned lastpos = i->points.back();
        unsigned lastx = lastpos % xSize;
        unsigned lasty = lastpos / xSize;
        
        if(lastx == x2 && lasty == y2)
        {
        	winner = *i; // got the real winner
        	break;
        }
        
        // Plot the point to demonstrate that we're processing it
        setColor(draw,4);
        draw.Point(lastx, lasty);
        draw.RefreshArea(lastx, lasty, 3,3);
        
        /* Create new paths out of this path. */
        list<path> newpaths;
        
        /* For all neighbours to this path's latest point */
        const set<unsigned> &neighbours = allneighbours[lastpos];
        for(set<unsigned>::const_iterator
            j = neighbours.begin();
            j != neighbours.end();
            ++j)
        {
            unsigned p = *j, nx = p % xSize, ny = p / xSize;
            
            double newcost = i->checkcost(nx, ny);
            
            // someone better already visited the point?
            if(seen[p] < newcost + THRESHOLD) continue;
            
            /* Visit the neighbour. This becomes a new path to analyze. */
            seen[p] = newcost;
            
            path newpath = *i;
            newpath.addpoint(nx, ny);

            newpaths.push_back(newpath);
            
            // Plot the point to demonstrate that we're considering it
            setColor(draw, 3);
            draw.Point(nx, ny);
            draw.RefreshArea(nx,ny, 3,3);
        }
        
        // Remove the path that was just analyzed.
        paths.erase(i);
        
        // Add the newly generated paths into the interesting list.
        paths.insert(newpaths.begin(), newpaths.end());
    }

    /* Delete the begin and end points from the list of points,
     * so that the next path we try to plot doesn't check these
     * points without need */
    DeleteNeighbour(x1,y1);
    DeleteNeighbour(x2,y2);
    
    return winner;
}

static void CalcRoute(unsigned x1,unsigned y1, unsigned x2,unsigned y2, DrawWindow& draw)
{
    /* Calculate the route from source to destination. */
    path winner = GetRoute(x1,y1, x2,y2, draw);
    
    /* Plot the route. */
    setColor(draw, 2);
    vector<unsigned>::const_iterator j, prev;
    for(j=prev=winner.points.begin(); j!=winner.points.end(); prev=j++)
    {
    	unsigned jx = *j    % xSize, jy = *j    / xSize;
    	unsigned lx = *prev % xSize, ly = *prev / xSize;
    	
        DrawLine(jx, jy, lx, ly, 2);
    }
}

static void drawCross(unsigned x, unsigned y, unsigned col)
{
    setColVal(x,y,col);
    setColVal(x-1,y-1,col); setColVal(x+1,y-1,col);
    setColVal(x-1,y+1,col); setColVal(x+1,y+1,col);
}

static void OnClick(DrawWindow& draw, unsigned x, unsigned y, unsigned)
{
    static unsigned sPointx=~0U, sPointy=~0U, ePointx=~0U, ePointy=~0U;
    static bool routecalculated = false;
    static bool edges_analyzed = false;

    if(sPointx==~0U) // source point not selected?
    {
        if(x>=xSize || y>=ySize) return;
        // set source point.
        sPointx = x; sPointy = y;
        drawCross(x,y,1);
        
    	if(!edges_analyzed)
    	{
            Draw(draw);
            fprintf(stderr, "Calculating edges (will only be done once)...\n");
            CalcNeighbours(draw);
            fprintf(stderr, "Done, now select the second point.\n");
            edges_analyzed=true;
    	}
    }
    else if(ePointx==~0U) // end point not selected?
    {
        if(x>=xSize || y>=ySize) return;
        if(x==sPointx && y==sPointy) return;
        // set end point.
        ePointx = x; ePointy = y;
        drawCross(x,y,1);
        fprintf(stderr, "Click button to find route.\n");
    }
    else if(!routecalculated) // find route if not done yet
    {
        CalcRoute(sPointx,sPointy,
                    ePointx,ePointy,
                    draw);
        for(unsigned i=0; i<xSize*ySize; i++)
            if(colorData[i]!=2) colorData[i]=0;
        routecalculated = true;
    }
    else // clear points and route
    {
        sPointx=sPointy=ePointx=ePointy=~0U;
        routecalculated=false;
        for(unsigned i=0; i<xSize*ySize; i++) colorData[i]=0;
    }
    
    // refresh window
    Draw(draw);
    draw.RefreshArea(0,0,xSize,ySize);
}

int main(int argc, char* argv[])
{
    if(argc<2)
    {
        cout << "Usage: " << argv[0] << " <tga image>" << endl; return 1;
    }

    unsigned char header[18];
    FILE* infile = fopen(argv[1], "rb");
    if(!infile)
    {
        cerr << "Can't open "; perror(argv[1]); return 1;
    }
    fread(header, 1, 18, infile);
    if(header[2]!=1 || header[16]!=8)
    {
        cerr << "Only uncompressed indexed 8-bit targas, thanks." << endl;
        return 1;
    }

    xSize = header[12]+header[13]*256;
    ySize = header[14]+header[15]*256;

    fseek(infile, header[0]+(header[5]+header[6]*256)*header[7]/8,
          SEEK_CUR);

    imageData.resize(xSize*ySize);
    colorData.resize(xSize*ySize, 0);

    unsigned char* image = new unsigned char[xSize*ySize];
    fread(image, 1, xSize*ySize, infile);
    for(unsigned i=0; i<xSize*ySize; i++)
        imageData[i] = (image[i]!=0);
    delete[] image;

    fclose(infile);

    DrawWindow draw(xSize, ySize, Draw, argc, argv);
    draw.OnClick(OnClick);
    draw.Done();
}
